"""
Functions which express problems in Sedumi format and export them as .mat files
for Matlab
"""

import os

import numpy as np
import scipy.io

from cvxopt import matrix as cvxmat


def write_cvxpy_to_mat(problem_data, target, simplify=True):
    '''
    Args:
        problem_data: As produced by applying get_problem_data['CVXOPT'] to a
        cvxpy problem.

    Returns: (None)

    Effect:
        Saves a .mat file containing the A, b, c, K that define the problem
        in Sedumi format to target (see http://plato.asu.edu/ftp/usrguide.pdf)
    '''

    A, b, c, K, offset = make_sedumi_format_problem(
        problem_data, simplify=simplify)
    assert offset == 0
    write_sedumi_to_mat(A, b, c, K, target)


def write_sedumi_to_mat(A, b, c, K, target):
    '''
    Args:
        A, b, c, K for Sedumi format
        target: the path where we will save the .mat

    Effect:
        Saves a .mat file containing A, b, c, K to target
    '''
    A = sparsify_tall_mat(A)
    b = sparsify_tall_mat(b)
    c = sparsify_tall_mat(c)
    K = clean_K_dims(K)

    # Check that target folder exists
    folder = os.path.dirname(target)
    if not os.path.exists(folder):
        os.makedirs(folder)

    scipy.io.savemat(target, {'A': A, 'b': b, 'c': c, 'K': K})


def clean_K_dims(K):
    '''
    Matlab requires the dimensions to be given in floating point numbers,
    this checker ensures that they are.
    '''
    for p in K:
        # let's try treating this like a list or tuple, and if that falls
        # through then we'll assume it's a single number instead.
        if isinstance(K[p], list):
            for i, x in enumerate(K[p]):
                K[p][i] = 1. * x
        else:
            K[p] = 1. * K[p]
    return K


def problem_data_prep(problem_data):
    '''
    'Touch up' the problem data in the following ways:
      - Make sure the matrix elements aren't integers
      - Make sure they're dense matrices rather than sparse ones, otherwise
        there seem to be difficulties constructing block matrices
      - Transpose c to be a row vector, which matches the organization of A, b, G, h
        (rows are for constraints, columns are for variables)
    '''
    problem_data['A'] = cvxmat(1. * problem_data['A'])
    problem_data['b'] = cvxmat(1. * problem_data['b'])
    problem_data['G'] = cvxmat(1. * problem_data['G'])
    problem_data['h'] = cvxmat(1. * problem_data['h'])
    problem_data['c'] = cvxmat(1. * problem_data['c']).T
    return problem_data


def make_sedumi_format_problem(problem_data, simplify=True):
    '''
    Input:
        problem_data: As produced by applying get_problem_data['CVXOPT'] to a
        cvxpy problem.
    Returns:
        A, b, c, K: Data defining an equivalent problem in Sedumi format.
    '''
    problem_data = problem_data_prep(problem_data)
    dims = problem_data['dims']
    assert not dims[
        'q'], "Sorry, at this time we can't handle SOC constraints!"

    nx = len(problem_data['c'])
    ni = dims['l']
    ne = len(problem_data['b'])
    num_sdp_vars = sum([s * s for s in dims['s']])

#==============================================================================
#   EXPANSION STEP:
#   Construct the expanded vector c_star and matrices A_star and Gs_star
#
#   At this point...
#     c_star = [ c ] (nx)
#              [ 0 ] (ni + num_sdp_vars)
#
#     A_star = [ A  |  0  |  0 ] (ne)        => eqs written in new vars
#              [ Gl |  I  |  0 ] (ni)        => ineqs written as eqs in new vars
#     Gs_star= [ Gs |  0  |  I ] (num_sdp_vars) => sets the exp in the SDP element
#               (nx) (ni)(num_sdp_vars)          equal to the var representing it
#==============================================================================
    num_sedumi_vars = nx + ni + num_sdp_vars

    c = np.zeros((1, num_sedumi_vars))
    c[0, 0:nx] = problem_data['c']

    A = np.zeros((ne + ni + num_sdp_vars, num_sedumi_vars))
    b = np.zeros((ne + ni + num_sdp_vars, 1))

    # Fill in blocks for Ax = b constraints
    A[0:ne, 0:nx] = problem_data['A']
    b[0:ne] = problem_data['b']

    # Fill in blocks for Gx + s = h
    A[ne:ne + ni, 0:nx] = problem_data['G'][:ni, :]  # = Gl
    A[ne:ne + ni, nx:nx + ni] = np.eye(ni)
    b[ne:ne + ni] = problem_data['h'][:ni, :]  # = hl

    # Fill out blocks defining h - Gs = vec(Y), where Y is the PSD matrix
    A[ne + ni:, 0:nx] = problem_data['G'][ni:, :]  # = Gs
    A[ne + ni:, nx + ni:] = np.eye(num_sdp_vars)
    b[ne + ni:] = problem_data['h'][ni:, :]  # = hs

    obj_cst = 0.
    K = {'f': nx, 'l': dims['l'], 'q': [], 's': dims['s']}
    if simplify:
        A, b, c, K, obj_cst = simplify_sedumi_model(A,
                                                    b,
                                                    c,
                                                    K,
                                                    allow_nonzero_b=False)
        assert obj_cst == 0, "This shouldn't be possible with allow_nonzero_b=False."
    else:
        A, b, c, K = symmetrize_sedumi_model(A, b, c, K)
    return A, b, c, K, obj_cst


def symmetrize_sedumi_model(A, b, c, K):
    '''
    Symmetrize sedumi model.
    '''
    colstart = K['f'] + K['l'] + sum(K['q'])

    for s in K['s']:
        for i in range(s):
            for j in range(i + 1, s):
                ijcol = colstart + i * s + j
                jicol = colstart + j * s + i
                averaged_Acol = 0.5 * (A[:, ijcol] + A[:, jicol])
                A[:, ijcol] = averaged_Acol
                A[:, jicol] = averaged_Acol
                averaged_c = 0.5 * (c[0, ijcol] + c[0, jicol])
                c[0, ijcol] = averaged_c
                c[0, jicol] = averaged_c
        colstart + s**2
    return A, b, c, K


def simplify_sedumi_model(A, b, c, K, allow_nonzero_b=False):
    '''
    Tries to eliminate variables using a few simple strategies:

    1. If a constraint is expressing :math:`A_{ki}x_i = b_k` where variable
    :math:`x_i` is a free variable, we can eliminate :math:`x_i`.

    2. If a constraint is expressing :math:`A_{ki}x_i + A_{kj}x_j = b_k`
    where variable :math:`x_i` is a free variable, we can eliminate
    :math:`x_i`.

    Args:
        A, b, c, K: for a problem in Sedumi format
        allow_nonzero_b: If False, only eliminate if bk = 0 is zero

    Returns:
        A, b, c, K: for the simplified problem.
        offset: A constant which must be added to the optimal value of the
        simplified problem in order to make it equivalent.  With
        allow_nonzero_b, offset will be 0.
    '''
    n_free = K['f']  # the first n_free variables will be eligible for any kind
    # of elimination
    n_nonneg = K['l']  # the next n_nonneg variables will be eligible for only
    # the simplest substitution aij*xj=bi and only in the case
    # where bi/aij >=0.
    n_vars = c.size
    n_ctr = b.size

#==============================================================================
#   SIMPLICATION STEP:
#   If some ctr k of A_star*x = hs is actually just equating xi = xj
#   for some i in the free vars, j in the sdp vars, we need to do the following:
#     (1) for A_star, c, add column/element i to column/element j
#     (2) for A_star, c, delete column/element i
#     (3) for A_star, delete row k
#     (4) adjust our counts for different variable/constraint types
#   On the first pass we will do (1) and make lists of which rows/ctrs to eliminate.
#   On a second pass we will do the rest.
#
#   SIMPLIFICATION PART ONE: Remove dependence on some cols and mark them for removal.
#==============================================================================
    offset = 0
    # Given var_i which is a free variable, figure out if there is a row k of
    # G_star such that Gs_star[ctr_k, var_i] == -1 AND hs[ctr_k] == 0 AND the
    # only other non-zero element in the row is Gs_star[ctr_k, nx + ni +
    # ctr_k] == 1
    for ctr_k in range(n_ctr):
        i, j = check_eliminatibility(A[ctr_k, :],
                                     b[ctr_k, 0],
                                     n_elig=n_free + n_nonneg,
                                     allow_nonzero_b=allow_nonzero_b)
        # Two cases where we can eliminate xi:
        # 1) xi is a free var
        free_ok = (i is not None and i < n_free)
        # 2) xi is a nonneg var, the ctr is of form Akixi = bk, and bk/Aki >= 0
        nonneg_ok = (i is not None and j is None and 1. *
                     b[ctr_k, 0] / A[ctr_k, i] >= 0)

        if free_ok or nonneg_ok:
            aki = A[ctr_k, i]
            bk = b[ctr_k, 0]
            factor = 1. * bk / aki

            # Akixi (optionally + Akjxj) = bk case, eliminate xi using
            # xi = (Akj/Aki) - (bk/Aki)*x_j
            b[:, 0] += -factor * A[:, i]
            offset += factor * c[0, i]

            if j is not None:
                # Akixi + Akjxj = bk case
                akj = A[ctr_k, j]
                factor = 1. * akj / aki
                A[:, j] += -factor * A[:, i]
                c[0, j] += -factor * c[0, i]

            # zero out the coefficients of var i to make sure it isn't chosen
            # for elimination again
            A[:, i] *= 0.
            c[0, i] *= 0.

    # To wrap up, list all the variables which are still nontrivial to the
    # model
    n_deleted_f = 0
    n_deleted_l = 0
    cols_to_keep = []
    vars_fl = n_free + n_nonneg
    for col in range(vars_fl):
        # free vars not in constraints must have 0 coeff in obj, else unbounded.
        # nonneg vars not in constraints must have >=0 coeff in obj, else unbounded.
        # if a var makes the probblem unbounded, we'll leave it alone and let
        # the user find out when they actually solve.
        free_and_deletable = col < n_free and c[0, col] == 0
        nneg_and_deletable = col >= n_free and col < vars_fl and c[0, col] >= 0
        if free_and_deletable and not abs(A[:, col]).any():
            n_deleted_f += 1
        elif nneg_and_deletable and not abs(A[:, col]).any():
            n_deleted_l += 1
        else:
            cols_to_keep.append(col)

    # SOC vars eliminatable iff they're not the first var of their vector and
    # they're unused in any ctrs.
    col_start = n_free + n_nonneg

    for i, q_size, in enumerate(K['q']):
        cols_to_keep.append(col_start)
        for col in range(col_start + 1, col_start + q_size):
            if c[0, col] == 0 and not abs(A[:, col]).any():
                K['q'][i] += -1
            else:
                cols_to_keep.append(col)
        col_start += q_size

    # All SDP vars kept
    cols_to_keep += range(col_start, n_vars)

    # Symmetrize the use of PSD matrix variables.  We do this now because it might
    # zero out some additional ctrs which we'll check for next.
    A, b, c, K = symmetrize_sedumi_model(A, b, c, K)

    # Now figure out which ctrs are nontrivial (trivial meaning 0x = 0).
    # Note that A ctr of 0x = b would make the problem infeasible, but in that case
    # we'll leave it in so the user finds it when they solve.
    rows_to_keep = []
    for row in range(n_ctr):
        if b[row, 0] != 0 or abs(A[row, :]).any():
            rows_to_keep.append(row)
#==============================================================================
#   SIMPLIFICATION STEP PART TWO: construct final matrices with only
#     the rows/cols we want
#==============================================================================
    # new downsized problem
    A = A[np.ix_(rows_to_keep, cols_to_keep)]
    b = b[np.ix_(rows_to_keep, [0])]
    c = c[np.ix_([0], cols_to_keep)]

    # problem dimensions
    assert len(cols_to_keep) + n_deleted_f + n_deleted_l
    K = {'f': K['f'] - n_deleted_f,
         'l': K['l'] - n_deleted_l,
         'q': K['q'],
         's': K['s']}
    return A, b, c, K, offset


def check_eliminatibility(g, h, n_elig=None, allow_nonzero_b=False):
    '''
    Tests if constraint :math:`gx = h` fits either pattern :math:`ax_i = d`
    or pattern :math:`ax_i + bx_j = d`, with the requirement that the
    :math:`x_i` variable be one of the first n_elig variables.

    Returns:
       ``i, None`` such that the constraint has the form :math:`ax_i = d` for
       some :math:`a, d`.

       ``i, j`` such that the constraint has the form :math:`ax_i + bx_j = d`
       for some :math:`a, b, d`.

       ``None, None`` if neither pattern applies.
    '''
    n = len(g)

    if n_elig is None:
        n_elig = n

    if not allow_nonzero_b and h != 0:
        return None, None

    for i in range(n_elig):
        if g[i] != 0:
            j = i + 1
            while j < n and g[j] == 0:
                j += 1
            # having stopped, see which we have found:
            if j == n:
                return i, None  # the end of the row
            elif not abs(g[j + 1:]).any():
                return i, j  # the second of exactly two nonzero coefficients
            else:
                return None, None  # the second of three or more nonzero coefficients

    # If nothing's been returned yet, it's because all the eliminatible
    # vars' coeffs are zero.
    return None, None


def sparsify_tall_mat(M, block_height=1000):
    '''
    Returns a sparse matrix in scipy.sparse.coo_matrix form which is equivalent to M
    '''
    i = 0
    spmat_collector = []
    while i * block_height < M.shape[0]:
        curr_block = M[i * block_height:(i + 1) * block_height, :]
        spmat_collector += [scipy.sparse.coo_matrix(curr_block.astype('d'))]
        i += 1
    return scipy.sparse.construct.vstack(spmat_collector)



"""

"""

import os.path

import sedumi_writer as sw
import solve_locally as ls
import result as res


MATLAB = 'matlab'
OCTAVE = 'octave'
NEOS = 'neos'


def check_output_target(mode, output_target):
    '''
    checks if the value of output_target is appropriate:
        - output_target can't be None if mode is matlab or octave
        - at least for now, output_target can't be a file that already exists
    '''
    assert mode in ['matlab', 'octave', 'neos'], \
        "Please choose mode equal to either 'matlab', 'octave', or 'neos'."
    assert mode not in ['matlab', 'octave'] or output_target, \
        "If mode is 'matlab' or 'octave', an output_target must be provided."
    if output_target:
        assert not os.path.exists(output_target), \
            ("Something already exists at output_target, we won't overwrite "
             "it:\n{0}".format(output_target))


def sdpt3_solve_problem(
        problem, mode, matfile_target,
        output_target=None, discard_matfile=True, **kwargs):
    '''
    A wrapper function that takes a cvxpy problem, makes the .mat file, solves
    it by NEOS or a local Matlab/SDPT3 installation, then constructs the result,
    prints it, and returns it.
    '''
    assert not os.path.exists(matfile_target), \
        ("Something already exists at matfile_target, we won't overwrite "
         "it:\n{0}".format(matfile_target))
    check_output_target(mode, output_target)

    # Write the problem to a .mat file in Sedumi format
    problem_data = problem.get_problem_data('CVXOPT')
    sw.write_cvxpy_to_mat(problem_data, matfile_target)

    return sdpt3_solve_mat(matfile_target,
                           mode,
                           output_target=output_target,
                           discard_matfile=discard_matfile,
                           **kwargs)


def sdpt3_solve_mat(
        matfile_path, mode, output_target=None, discard_matfile=True, **kwargs):
    '''
    A wrapper function that takes the path of a .mat file, solves the Sedumi
    problem it contains with NEOS or a local Matlab/SDPT3 installation, then
    constructs the result, prints it, and returns it.
    '''
    matfile_path = os.path.abspath(matfile_path)
    check_output_target(mode, output_target)

    # Depending on the mode, solve the problem using a local Matlab+SDPT3
    # installation or on the NEOS server
    if mode == MATLAB:
        msg = ls.matlab_solve(matfile_path,
                              discard_matfile=discard_matfile)
    elif mode == OCTAVE:
        msg = ls.octave_solve(matfile_path,
                              discard_matfile=discard_matfile,
                              **kwargs)
    elif mode == NEOS:
        import solve_neos as ns
        msg = ns.neos_solve(matfile_path,
                            discard_matfile=discard_matfile)

    if output_target:
        with open(output_target, "w") as fp:
            fp.write(msg)

    result = res.make_result_dict(msg)
    return result



"""

"""

import re
from numpy import zeros


_SDPT3_POS_STATUS_MAP_VERB = (
    'max(relative gap,infeasibility) < gaptol (OPTIMAL)',
    'primal problem is suspected to be infeasible',
    'dual problem is suspected to be infeasible',
    'norm(X) or norm(Z) diverging'
)

_SDPT3_NEG_STATUS_MAP_VERB = (
    'max(relative gap,infeasibility) < gaptol (OPTIMAL)',
    'relative gap < infeasibility',
    'lack of progress in predictor or corrector',
    'X or Z not positive definite',
    'difficulty in computing predictor or corrector direction',
    'progress in relative gap or infeasibility is bad',
    'maximum number of iterations reached',
    'primal infeasibility has deteriorated too much',
    'progress in relative gap has deteriorated',
    'lack of progress in infeasibility'
)


def make_result_dict(msg):
    '''
    Extracts some solve information from the log message and constructs a dict.
    This function will error if the message is empty or does not include the phrase
    "SDPT3: Infeasible path-following algorithms", which is an indicator that the
    solver at least started okay.  If the log passes that basic test, we just retrieve
    what information we can.  If the dict doesn't at least contain a status_num and
    status_verb, you should check the log manually and see what went wrong.
    '''
    assert can_use_msg(
        msg), "Stopping, the message is not properly formed: " + msg
    result_dict = extract_prop_dict(msg)
    result_dict['Xvars'] = extract_X(msg)
    result_dict['status_verb'] = get_verb_status(result_dict['status_num'])
    result_dict['msg'] = msg
    return result_dict


def make_result_summary(result):
    '''
    Prints a basic summary of information about an SDPT3 solve result.
    '''
    return (
        "SDPT3 solve finished with status code {0[status_num]}: {0[status_verb]}"
        "Primal value (from X): {0[primal_z]} (infeasibility: {0[rel_primal_feas]})"
        "Dual value (bound): {0[dual_z]} (infeasibility: {0[rel_dual_feas]})"
        "Relative gap: {0[rel_gap]}"
        "Solve time: {1} s"
    ).format(result, 0.001 * round(1000 * result['solve_time']))


def print_summary(result):
    ''' Prints out the summary produced by make_result_summary. '''
    print "\nResult summary:\n" + make_result_summary(result)


def can_use_msg(msg):
    '''
    Checks to see if a message is non-trivial and contains the phrase
    "SDPT3: Infeasible path-following algorithms", which should appear in the
    output log just before the result properties are listed.  While it's
    probably possible for this function to return True in a case where there's
    something wrong with the message, this is a good preliminary check to do.
    '''
    return msg is not None and "SDPT3: Infeasible path-following algorithms" in msg


def extract_prop_dict(msg):
    '''
    Given the output message from running SDPT3solve.m, this function
    constructs and returns a dictionary of basic solve result information.
    '''
    result_dict = {}

    # the key is what we look for in the SDPT3 message output, the value is
    # the name of key we'll save the information to in the output dict.
    keylist = {'number of iterations': 'iterations',
               'primal objective value': 'primal_z',
               'dual objective value': 'dual_z',
               'gap': 'abs_gap',
               'relative gap': 'rel_gap',
               'actual relative gap': 'actual_rel_gap',
               'rel. primal infeas': 'rel_primal_feas',
               'rel. dual infeas': 'rel_dual_feas',
               'Total CPU time (secs)': 'solve_time',
               'CPU time per iteration': 'solve_time_per_iter',
               'termination code': 'status_num'}

    for key in keylist.values():
        result_dict[key] = None

    line_pattern = re.compile(
        r'\w[ \w:=.()]*[ \<\>]*=[ <>]*[ \d\.\-+e]*[\d\.\-+e]')
    phrase_pattern = re.compile(r'[\w\d\.\-+()][ \w\d\.\-+()]*')

    line_list = line_pattern.findall(msg)
    for line in line_list:
        line_parts = phrase_pattern.findall(line)
        key = line_parts[0].rstrip().replace("dual  ", "dual")
        val = handle_msg_item(line_parts[-1].strip())

        if key in keylist and keylist[key] is not None:
            result_dict[keylist[key]] = val
    return result_dict


def extract_X(msg):
    '''
    Given the output message from running SDPT3solve.m, reconstruct the X
    matrix from the printed output and return it.
    '''
    # First pull out the definitions.  Each starts with 'X{numbers} ='
    # and we'll keep grabbing text until we hit any of the characters >, <, *,
    # or X
    var_pattern = re.compile(r'X\{[\d]*\} =[^><\*X]*')
    var_list = var_pattern.findall(msg)

    # Xlist will hold the solution variables or matrices
    Xlist = [None] * len(var_list)

    for i, Xmsg in enumerate(var_list):
        # Drop the "X{something} =" and strip extra whitespace
        Xstart = Xmsg.find('=') + 1
        Xmsg = Xmsg[Xstart:].strip()

        # From Xmsg we can judge if the current variable is a matrix displaying in chunks
        # due to display width constraints, a matrix displaying all in one chunk,
        # or a scalar.
        if 'Column' in Xmsg:
            # If X[i] is a chunked matrix, each chunk has a header containing 'Column'
            # Split the data into chunks
            chunk_pattern = re.compile('Column[^C]*')
            chunk_list = chunk_pattern.findall(Xmsg)

            # Use regular expressions to split the data into header parts and
            # data parts
            header = re.compile(r'Columns* [\d]+[ through [\d]+]?')
            header_list = header.findall(Xmsg)
            chunk_list = header.split(Xmsg)[1:]  # drop the first, which is ''
            assert len(header_list) == len(chunk_list)

            # Find out how many columns the matrix has by grabbing the last column number
            # from the last header
            int_pattern = re.compile(r'\d[\d]*')
            int_list = int_pattern.findall(header_list[-1])
            cols = int(int_list[-1])
            # We'll count rows and initialize the matrix during the processing of
            # the first chunk.

            # Take a matching header and chunk and place the chunk's data in
            # the columns of Xlist[i] which are indicated by the header.
            for header, chunk in zip(header_list, chunk_list):
                # Grab the column numbers from the header, note the first one. It's
                # the start column we'll use to place the chunk's data in the
                # matrix.
                int_list = [int(x) for x in int_pattern.findall(header)]
                col_start = int_list[0] - 1

                # Split the chunk into rows
                chunk = chunk.strip()
                chunk = re.sub(' *\n *', '\n', chunk)
                chunk_lines = chunk.split('\n')

                # Initialize the matrix now if it hasn't been done yet.
                if Xlist[i] is None:
                    rows = len(chunk_lines)
                    Xlist[i] = zeros((rows, cols))

                # Plug the row's data into the matrix
                for row, line in enumerate(chunk_lines):
                    for k, item in enumerate(re.split(r'\s+', line)):
                        Xlist[i][row, col_start + k] = float(item)
            print "Imported X[{0}] as a matrix with shape {1}.".format(i, Xlist[i].shape)

        elif ' ' in Xmsg or '\n' in Xmsg:
            # Otherwise if it has spaces or line breaks it's a non-chunked
            # matrix
            Xmsg = Xmsg.strip()
            Xmsg = re.sub(' +', ' ', Xmsg)
            Xmsg = re.sub(' *\n *', '\n', Xmsg)
            Xmsg_lines = Xmsg.split('\n')

            # Initialize the matrix.
            rows = len(Xmsg_lines)
            cols = len(Xmsg_lines[0].split())
            Xlist[i] = zeros((rows, cols))

            # Plug the row's data into the matrix
            for row, line in enumerate(Xmsg_lines):
                for k, item in enumerate(re.split(r'\s+', line.strip())):
                    Xlist[i][row, k] = float(item)
            print "Imported X[{0}] as a matrix with shape {1}.".format(i, Xlist[i].shape)

        else:
            # Otherwise it's a scalar
            Xlist[i] = float(Xmsg)
            print "Imported X[{0}] as a scalar.".format(i)

    return Xlist


def handle_msg_item(x):
    '''
    A function that takes a string x and returns it's interpretation as an int,
    float, or string in that order of preference.  If x is None, None is
    returned.
    '''
    if len(x) == 0:
        return None
    else:
        try:
            x = float(x)
            if int(x) == x:
                return int(x)
            else:
                return x
        except (TypeError, ValueError):
            pass
    return x


def get_verb_status(status_num):
    '''
    A function that takes an SDPT3 numerical termination code as input and
    returns a phrase (string) explaining the implications of the termination
    code.
    '''
    if status_num is None:
        return "no termination code given"
    elif status_num >= 0:
        return _SDPT3_POS_STATUS_MAP_VERB[status_num]
    else:
        return _SDPT3_NEG_STATUS_MAP_VERB[-status_num]



"""

"""

from solve import check_output_target
from solve import sdpt3_solve_problem
from solve import sdpt3_solve_mat
from solve import MATLAB
from solve import OCTAVE
from solve import NEOS
from sedumi_writer import write_cvxpy_to_mat
from sedumi_writer import write_sedumi_to_mat
from result import print_summary



""" sdpt3glue package information.
"""
from setuptools import setup, find_packages


def _load_requires_from_file(filepath):
    """ Read a package list from a given file path.

    Args:
      filepath: file path of the package list.

    Returns:
      a list of package names.
    """
    with open(filepath) as fp:
        return [pkg_name.strip() for pkg_name in fp.readlines()]


setup(
    name='sdpt3glue',
    version='0.9.1',
    description=(
        'Glue code for solving semidefinite programs '
        'in Cvxpy format using the SDPT3 package for Matlab.'
    ),
    author="Trish Gillett-Kawamoto",
    author_email="discardthree@gmail.com",
    url="https://github.com/discardthree/PySDPT3glue",
    packages=find_packages(exclude=["tests"]),
    package_data={'sdpt3glue': ['*.m']},
    install_requires=_load_requires_from_file("requirements.txt"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Mathematics"
    ],
    test_suite='tests.suite',
    license="MIT"
)



'''
Methods that run code on a copy of Matlab or Octave installed on the user's
machine.
'''

import os
import os.path

import shutil
import subprocess
import tempfile


class MatlabCallError(Exception):
    '''
    This error is raised when an error occurs during a subprocess call to Matlab.
    '''
    pass


class OctaveCallError(Exception):
    '''
    This error is raised when an error occurs during a subprocess call to Octave.
    '''
    pass


class SubprocessCallError(Exception):
    '''
    This error is raised when an error occurs during a subprocess call.
    '''
    pass


def matlab_solve(matfile_target, discard_matfile=True):
    '''
    The .mat is loaded into matlab and the problem is solved with SDPT3.

    Args:
        matfile_target: the path to the .mat file containing the Sedumi format problem data.
        discard_matfile: if True, deletes the .mat file after the solve finishes.

    Returns:
        A dictionary with solve result information.
    '''
    # Generating the .mat file
    run_command = "matlab -r \"SDPT3solve('{0}')\" -nodisplay -nojvm".format(
        matfile_target)
    try:
        msg = run_command_get_output(run_command)
    except SubprocessCallError, e:
        raise MatlabCallError(e)

    # Cleanup
    if discard_matfile:
        print "now deleting {0}".format(matfile_target)
        os.remove(matfile_target)

    return msg


def octave_solve(matfile_target, discard_matfile=True, cmd="octave"):
    '''
    The .mat is loaded into octave and the problem is solved with SDPT3.

    Args:
        matfile_target: the path to the .mat file containing the Sedumi format problem data.
        discard_matfile: if True, deletes the .mat file after the solve finishes.
        cmd: command name for octave, which will be used for alternative command.

    Returns:
        A dictionary with solve result information.
    '''
    # Generating the .mat file
    with tempfile.NamedTemporaryFile(
        suffix=".m", dir=os.path.dirname(matfile_target)) as runner:

        with open(os.path.join(os.path.dirname(__file__), "SDPT3solve.m")) as lib:
            shutil.copyfileobj(lib, runner)

        runner.write("SDPT3solve('{0}');\n".format(
            os.path.relpath(matfile_target)))
        runner.flush()

        run_command = "{cmd} {script}".format(
            cmd=cmd, script=os.path.relpath(runner.name))
        try:
            msg = run_command_get_output(run_command)
        except SubprocessCallError, e:
            raise OctaveCallError(e)

    # Cleanup
    if discard_matfile:
        print "now deleting {0}".format(matfile_target)
        os.remove(matfile_target)

    return msg


def run_command_get_output(run_command):
    '''
    Runs the command run_command, saves the output to output_target, and
    returns the output log.
    '''
    # Performing the Matlab solve
    try:
        proc = subprocess.Popen(
            run_command, shell=True, stdout=subprocess.PIPE,
            cwd=os.path.join(os.path.dirname(__file__)))
        return proc.communicate()[0]

    except:
        err_msg = ("Something went wrong with the command. The command "
                   "run was:\n{0}\nPlease check that the command looks good "
                   "and that the the folder containing SDPT3solve.m is in the "
                   "Matlab or Octave path variable.").format(run_command)
        raise SubprocessCallError(err_msg)



"""
Basic scraper/webdriver to submit an sdpt3 problem on the NEOS webpage.
This version uses selenium because difficulties were encountered using XML-RPC
for problems of this type.
"""

import os
import sys
import contextlib
from time import sleep
import xmlrpclib
from xmlrpclib import Fault
from xmlrpclib import ProtocolError
from selenium import webdriver
from selenium.common.exceptions import WebDriverException


WEBDRIVERS = {
    "Firefox": webdriver.Firefox,
    "PhantomJS": webdriver.PhantomJS
}
""" Webdrivers to try using. """


def _print_fault(err, space=2):
    """ Print message from xmlrpclib.Fault.

    Args:
      err: xmlrpclib.Fault error.
      space: Space size to be added before each message.
    """
    print " " * space + "A fault occurred"
    print " " * space + "Fault code: {0}".format(err.faultCode)
    print " " * space + "Fault string: {0}".format(err.faultString)


def _print_protocol_error(err, space=2):
    """ Print message from xmlrpclib.ProtocolError.

    Args:
      err: xmlrpclib.ProtocolError error.
      space: Space size to be added before each message.
    """
    print " " * space + "A protocol error occurred"
    print " " * space + "URL: {0}".format(err.url)
    print " " * space + "HTTP/HTTPS headers: {0}".format(err.headers)
    print " " * space + "Error code: {0}".format(err.errcode)
    print " " * space + "Error message: {0}".format(err.errmsg)


@contextlib.contextmanager
def _get_driver():
    """ Try to make a webdriver according to installed browsers.

    Returns:
      Installed webdriver.

    Raises:
      RuntimeError: When no supported browsers are available.
    """
    for name, driver in WEBDRIVERS.items():
        try:
            res = driver()

        except (StandardError, WebDriverException) as e:
            sys.stderr.write(
                "{0} seems to be not installed: {1}\n".format(name, e))

        else:
            try:
                yield res
            finally:
                res.close()
                res.quit()
            return

    raise WebDriverException("No web drivers are available.")


class NeosError(Exception):
    """ Error caused by Neos server.

    This error is raised when an error occurs during calling Neos.
    """
    pass


def neos_solve(
        matfile_target, discard_matfile=True, no_prompt=False,
        return_id_pwd=False):
    '''
    Submits the Sedumi format .mat file to be solved on NEOS with SDPT3.
    Returns the solve result message from NEOS.

    If no_prompt is True, it won't ask for id and password manually and raise
    NeosError when some errors occur during connection to neos server.

    Raises:
      NeosError: When an error occurs by using Neos server.
    '''
    jobid, pwd = handle_submission(matfile_target, no_prompt=no_prompt)

    neos_int = NeosInterface()
    msg = neos_int.track_and_return(jobid, pwd)

    # Cleanup
    if discard_matfile:
        os.remove(matfile_target)

    if return_id_pwd:
        return msg, jobid, pwd
    else:
        return msg


def handle_submission(matfile_target, no_prompt=False):
    '''
    Submits the Sedumi format .mat file to be solved on NEOS with SDPT3.

    Args:
       matfile_target: the path of the .mat file containing the problem data.

    Returns:
       no_prompt: If False, in the event that problem submission fails the
       user may be prompted to manually submit the problem on the website and
       copy-paste the job ID and password information into a prompt.  If True,
       an error will thrown rather than prompting occurring.  Make sure to use
       True if running headlessly or otherwise unmonitored.

    Raises:
      NeosError: When an error occurs by using Neos server.
    '''
    if not os.path.exists(matfile_target):
        raise ValueError(
            "The matfile {0} doesn't exist".format(matfile_target))

    # any backslashes need to be doubly escaped for the web form
    matfile_target = matfile_target.replace('\\', '\\\\')

    try:
        with _get_driver() as browser:
            browser.get(
                'http://www.neos-server.org/neos/solvers/sdp:sdpt3/MATLAB_BINARY.html')

            # Find the .mat upload box and input the path to ours
            file_upload_element = browser.find_element_by_name("field.2")
            file_upload_element.send_keys(matfile_target)
            assert file_upload_element.get_attribute('value'), \
                "Couldn't input file name, are you sure it exists?"

            # Find the submit button and click it
            submit_xpath = '//input[@type="submit"]'
            submit_button = browser.find_element_by_xpath(submit_xpath)
            submit_button.click()

            # After a few moments, the id and password will appear, and
            # we'll try to grab them automatically.
            source = browser.page_source
            jobid, pwd = extract_id_pwd(source)

    except WebDriverException as e:
        if no_prompt:
            raise NeosError(e)

        # If that fails for any reason, we ask the user to submit the problem
        # manually and copy-paste the lines giving the id and password.
        try:
            jobid, pwd = ask_user_to_submit(matfile_target)

        except EOFError as e:
            raise NeosError(e)

    return jobid, pwd


def extract_id_pwd(source):
    '''
    Given a snippet of the form

    .. code-block:: none

        Job#     : xxxxxxx
        Password : yyyyyyyy

    extracts and returns the strings for job ID and password

    '''
    start = source.find('Job#')
    snippet = source[start:start + 60]
    lines = snippet.split('\n')

    jobid_line = lines[0].strip().split()
    pwd_line = lines[1].strip().split()

    jobid = int(jobid_line[-1])
    pwd = pwd_line[-1]
    return jobid, pwd


def ask_user_to_submit(matfilepath):
    '''
    Instructs the user to manually submit their problem and then feed the ID
    and password back into the program.

    Raises:
      EOFError: When a user sends EOF.
    '''
    print (
        "Selenium submission failed, submit it manually!"
        "Find this files and submit them on the website:"
        "{0}"
        "Once you've submitted your problem, copy and paste the"
        "two lines that look like this"
        "        Job#     : xxxxxxx"
        "        Password : yyyyyyyy"
        "below and hit Enter. Or, you can just enter the job and"
        "password strings on consecutive lines.  Good luck!"
    ).format(matfilepath)

    user_input = ''
    while True:

        while not user_input:
            user_input = raw_input()

        try:
            jobid = int(user_input.strip().split()[-1])
            pwd = raw_input().strip().split()[-1]
            print "\n=============\n"
            return jobid, pwd

        except IndexError:
            print "\nTry again!\n"


class NeosInterface(object):
    """
    An abstract class for connections with the remote NEOS Server for
    Optimization. NeosInterface objects communicate with the NEOS Server via
    XML-RPC. This class wraps the server's services into a few convenient
    methods which, for the time being, are designed with the solution of AMPL
    models in mind.
    """

    def __init__(self, neos_host=None, neos_port=None):
         # Go to xmlrpc and flip on __verbose if you are debugging
         # and want to see ALL xml-rpc communication
        if not neos_host:
            neos_host = "neos.mcs.anl.gov"
        if not neos_port:
            neos_port = 3332

        neos_url = "http://{host}:{port}".format(
            host=neos_host, port=neos_port)
        self.server = xmlrpclib.ServerProxy(neos_url)

    def track_and_return(self, jobid, pwd):
        '''
        Takes a jobid and password for a solve already in progress,
        waits for its status to change to 'Done', and returns the message,
        jobid, and password.
        '''
        k = 20
        while True:

            try:
                status = self.server.getJobStatus(jobid, pwd)

            except Fault as err:
                _print_fault(err)
                print "  Error checking status: {0}  (will try again)".format(
                    sys.exc_info()[0])

            except ProtocolError as err:
                _print_protocol_error(err)
                print "  Error checking status: {0}  (will try again)".format(
                    sys.exc_info()[0])

            else:
                if status == "Done":
                    break

            sleep(k)

        msg = self.get_final_results(jobid, pwd).data
        return msg

    def get_final_results(self, jobid, pwd):
        '''
        This funtion is 'stubborn', meaning that if it's unable to get the
        results it will wait a few seconds and try again.  Stubborn
        functionality is intended to smooth over lost internet connections or
        computers that may fall asleep and then wake up and resume running
        code.
        '''
        while True:
            try:
                return self.server.getFinalResults(jobid, pwd)
            except Fault as err:
                _print_fault(err)
            except ProtocolError as err:
                _print_protocol_error(err)

            sleep(3)
